/* cfun-list.h
 *
 * COPYRIGHT (c) 1994 AT&T Bell Laboratories.
 *
 * This file lists the directory library of C functions that are callable by ML.
 */

#ifndef CLIB_NAME
#define CLIB_NAME	"<<your library name here>>"
#define CLIB_VERSION	"<<version tag>>"
#define CLIB_DATE	"<<version date>>"
#endif

CFUNC("<<bind name>>",	<<C name>>,		"<<ML type (optional)>>")

