/* standalone.c
 *
 * COPYRIGHT (c) 1993 by AT&T Bell Laboratories.
 *
 * This is the main routine for linking with stand-alone applications.
 */

#include "ml-base.h"
#include "ml-limits.h"

main (argc, **argv)
    int		arg;
    char	**argv;
{

} /* end of main. */

